@@ -14,6 +14,7 @@
 #include "btree/write_buffer.h"
 
 #include "data/compress.h"
+#include "data/copygc.h"
 #include "data/ec.h"
 #include "data/move.h"
 #include "data/reconcile.h"
@@ -295,9 +296,7 @@ static inline struct bch_backpointer rb_bp(enum btree_id btree, unsigned level,
 static int reconcile_bp_del(struct btree_trans *trans, enum btree_id btree, unsigned level,
 			    struct bkey_s_c k, struct bpos bp_pos)
 {
-	CLASS(btree_iter, iter)(trans, BTREE_ID_reconcile_scan, bp_pos,
-				BTREE_ITER_intent|
-				BTREE_ITER_with_updates);
+	CLASS(btree_iter, iter)(trans, BTREE_ID_reconcile_scan, bp_pos, BTREE_ITER_intent);
 	struct bkey_s_c bp_k = bkey_try(bch2_btree_iter_peek_slot(&iter));
 
 	struct bch_backpointer bp = rb_bp(btree, level, k);
@@ -546,11 +545,12 @@ static inline bool bkey_should_have_rb_opts(struct bkey_s_c k,
 	return new.need_rb;
 }
 
-static bool bch2_bkey_needs_reconcile(struct bch_fs *c, struct bkey_s_c k,
-				      struct bch_inode_opts *opts,
-				      int *need_update_invalid_devs,
-				      struct bch_extent_reconcile *ret)
+static int bch2_bkey_needs_reconcile(struct btree_trans *trans, struct bkey_s_c k,
+				     struct bch_inode_opts *opts,
+				     int *need_update_invalid_devs,
+				     struct bch_extent_reconcile *ret)
 {
+	struct bch_fs *c = trans->c;
 	bool btree = bkey_is_btree_ptr(k.k);
 
 	if (btree &&
@@ -575,56 +575,58 @@ static bool bch2_bkey_needs_reconcile(struct bch_fs *c, struct bkey_s_c k,
 	bool incompressible = false, unwritten = false, ec = false;
 	unsigned durability = 0, durability_acct = 0, invalid = 0, min_durability = INT_MAX;
 
-	scoped_guard(rcu) {
-		const union bch_extent_entry *entry;
-		struct extent_ptr_decoded p;
-		unsigned ptr_bit = 1;
+	const union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+	unsigned ptr_bit = 1;
 
-		bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
-			incompressible	|= p.crc.compression_type == BCH_COMPRESSION_TYPE_incompressible;
-			unwritten	|= p.ptr.unwritten;
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
+		incompressible	|= p.crc.compression_type == BCH_COMPRESSION_TYPE_incompressible;
+		unwritten	|= p.ptr.unwritten;
 
-			struct bch_dev *ca = bch2_dev_rcu_noerror(c, p.ptr.dev);
-			if (ca && !p.ptr.cached) {
-				if (!poisoned &&
-				    !btree &&
-				    p.crc.csum_type != csum_type)
-					r.need_rb |= BIT(BCH_REBALANCE_data_checksum);
+		bool evacuating = bch2_dev_bad_or_evacuating(c, p.ptr.dev);
 
-				if (!poisoned &&
-				    p.crc.compression_type != compression_type)
-					r.need_rb |= BIT(BCH_REBALANCE_background_compression);
+		if (!poisoned &&
+		    !btree &&
+		    p.crc.csum_type != csum_type)
+			r.need_rb |= BIT(BCH_REBALANCE_data_checksum);
 
-				if (!poisoned &&
-				    r.background_target &&
-				    !bch2_dev_in_target(c, p.ptr.dev, r.background_target)) {
-					r.need_rb |= BIT(BCH_REBALANCE_background_target);
-					r.ptrs_moving |= ptr_bit;
-				}
+		if (!poisoned &&
+		    p.crc.compression_type != compression_type)
+			r.need_rb |= BIT(BCH_REBALANCE_background_compression);
 
-				if (ca->mi.state == BCH_MEMBER_STATE_evacuating) {
-					r.need_rb |= BIT(BCH_REBALANCE_data_replicas);
-					r.hipri = 1;
-					r.ptrs_moving |= ptr_bit;
-				}
-
-				unsigned d = __extent_ptr_durability(ca, &p);
-
-				durability_acct += d;
-
-				if (ca->mi.state == BCH_MEMBER_STATE_evacuating)
-					d = 0;
-
-				durability += d;
-				min_durability = min(min_durability, d);
-
-				ec |= p.has_ec;
-			}
-
-			invalid += p.ptr.dev == BCH_SB_MEMBER_INVALID;
-
-			ptr_bit <<= 1;
+		if (!poisoned &&
+		    !evacuating &&
+		    !p.ptr.cached &&
+		    r.background_target &&
+		    !bch2_dev_in_target(c, p.ptr.dev, r.background_target)) {
+			r.need_rb |= BIT(BCH_REBALANCE_background_target);
+			r.ptrs_moving |= ptr_bit;
 		}
+
+		if (evacuating) {
+			r.need_rb |= BIT(BCH_REBALANCE_data_replicas);
+			r.hipri = 1;
+			r.ptrs_moving |= ptr_bit;
+		}
+
+		int d = bch2_extent_ptr_desired_durability(trans, &p);
+		if (d < 0)
+			return d;
+
+		durability_acct += d;
+
+		if (evacuating)
+			d = 0;
+
+		durability += d;
+		if (!p.ptr.cached)
+			min_durability = min(min_durability, d);
+
+		ec |= p.has_ec;
+
+		invalid += p.ptr.dev == BCH_SB_MEMBER_INVALID;
+
+		ptr_bit <<= 1;
 	}
 
 	if (unwritten || incompressible)
@@ -852,8 +854,9 @@ int bch2_bkey_set_needs_reconcile(struct btree_trans *trans,
 	int need_update_invalid_devs;
 	struct bch_extent_reconcile new;
 
-	if (!bch2_bkey_needs_reconcile(c, k.s_c, opts, &need_update_invalid_devs, &new))
-		return 0;
+	int ret = bch2_bkey_needs_reconcile(trans, k.s_c, opts, &need_update_invalid_devs, &new);
+	if (ret <= 0)
+		return ret;
 
 	struct bch_extent_reconcile *old =
 		(struct bch_extent_reconcile *) bch2_bkey_reconcile_opts(c, k.s_c);
@@ -875,8 +878,9 @@ int bch2_bkey_set_needs_reconcile(struct btree_trans *trans,
 
 	if (unlikely(need_update_invalid_devs)) {
 		if (need_update_invalid_devs > 0) {
-			bch2_bkey_drop_ptrs(k, p, entry,
+			bch2_bkey_drop_ptrs_noerror(k, p, entry,
 				(p.ptr.dev == BCH_SB_MEMBER_INVALID &&
+				 !p.has_ec &&
 				 need_update_invalid_devs &&
 				 need_update_invalid_devs--));
 		} else {
@@ -919,8 +923,9 @@ int bch2_update_reconcile_opts(struct btree_trans *trans,
 	int need_update_invalid_devs;
 	struct bch_extent_reconcile new;
 
-	if (!bch2_bkey_needs_reconcile(c, k, opts, &need_update_invalid_devs, &new))
-		return 0;
+	int ret = bch2_bkey_needs_reconcile(trans, k, opts, &need_update_invalid_devs, &new);
+	if (ret <= 0)
+		return ret;
 
 	if (!level) {
 		struct bkey_i *n = errptr_try(bch2_trans_kmalloc(trans, bkey_bytes(k.k) +
@@ -1271,24 +1276,29 @@ static int reconcile_set_data_opts(struct btree_trans *trans,
 	unsigned compression_type = bch2_compression_opt_to_type(r->background_compression);
 
 	if (r->need_rb & BIT(BCH_REBALANCE_data_replicas)) {
-		unsigned durability = bch2_bkey_durability(c, k);
+		int durability = bch2_bkey_durability(trans, k);
+		if (durability < 0)
+			return durability;
+
 		unsigned ptr_bit = 1;
 
-		guard(rcu)();
 		if (durability <= r->data_replicas) {
+			guard(rcu)();
+
 			bkey_for_each_ptr(ptrs, ptr) {
-				struct bch_dev *ca = bch2_dev_rcu_noerror(c, ptr->dev);
-				if (ca && !ptr->cached && !ca->mi.durability)
+				if (bch2_dev_bad_or_evacuating(c, ptr->dev))
 					data_opts->ptrs_kill |= ptr_bit;
 				ptr_bit <<= 1;
 			}
-
-			data_opts->extra_replicas = r->data_replicas - durability;
 		} else {
 			bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
-				unsigned d = bch2_extent_ptr_durability(c, &p);
+				int d = bch2_extent_ptr_durability(trans, &p);
+				if (d < 0)
+					return d;
 
-				if (d && durability - d >= r->data_replicas) {
+				if (bch2_dev_bad_or_evacuating(c, p.ptr.dev) ||
+				    (!p.ptr.cached &&
+				     d && durability - d >= r->data_replicas)) {
 					data_opts->ptrs_kill |= ptr_bit;
 					durability -= d;
 				}
@@ -1314,14 +1324,16 @@ static int reconcile_set_data_opts(struct btree_trans *trans,
 			if (extent_ec_pending(trans, ptrs))
 				return false;
 
-			data_opts->extra_replicas = r->data_replicas;
+			data_opts->extra_replicas = 1;
+			data_opts->no_devs_have = true;
+
+			if (r->need_rb == BIT(BCH_REBALANCE_erasure_code))
+				data_opts->write_flags |= BCH_WRITE_must_ec;
 		} else {
 			unsigned ptr_bit = 1;
 			bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
-				if (p.has_ec) {
+				if (p.has_ec)
 					data_opts->ptrs_kill_ec |= ptr_bit;
-					data_opts->extra_replicas += p.ec.redundancy;
-				}
 
 				ptr_bit <<= 1;
 			}
@@ -1333,29 +1345,28 @@ static int reconcile_set_data_opts(struct btree_trans *trans,
 		bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
 			if ((r->need_rb & BIT(BCH_REBALANCE_data_checksum)) &&
 			    p.crc.csum_type != csum_type)
-				data_opts->ptrs_rewrite |= ptr_bit;
+				data_opts->ptrs_kill |= ptr_bit;
 
 			if ((r->need_rb & BIT(BCH_REBALANCE_background_compression)) &&
 			    p.crc.compression_type != compression_type)
-				data_opts->ptrs_rewrite |= ptr_bit;
+				data_opts->ptrs_kill |= ptr_bit;
 
 			if ((r->need_rb & BIT(BCH_REBALANCE_background_target)) &&
-			    !bch2_dev_in_target(c, p.ptr.dev, r->background_target))
-				data_opts->ptrs_rewrite |= ptr_bit;
+			    !p.ptr.cached &&
+			    !bch2_dev_in_target_rcu(c, p.ptr.dev, r->background_target))
+				data_opts->ptrs_kill |= ptr_bit;
 
 			ptr_bit <<= 1;
 		}
 	}
 
-	bool ret = (data_opts->ptrs_rewrite ||
-		    data_opts->ptrs_kill ||
+	bool ret = (data_opts->ptrs_kill ||
 		    data_opts->ptrs_kill_ec ||
 		    data_opts->extra_replicas);
 	if (!ret) {
-		CLASS(printbuf, buf)();
-		prt_printf(&buf, "got extent to reconcile but nothing to do, confused\n  ");
-		bch2_bkey_val_to_text(&buf, c, k);
-		bch_err(c, "%s", buf.buf);
+		CLASS(bch_log_msg_ratelimited, msg)(c);
+		prt_printf(&msg.m, "got extent to reconcile but nothing to do, confused\n  ");
+		bch2_bkey_val_to_text(&msg.m, c, k);
 	}
 
 	return ret;
@@ -1371,7 +1382,7 @@ static void bkey_reconcile_pending_mod(struct bch_fs *c, struct bkey_i *k, bool
 }
 
 static int bch2_extent_reconcile_pending_mod(struct btree_trans *trans, struct btree_iter *iter,
-					     struct bkey_s_c k, bool set)
+					     unsigned level, struct bkey_s_c k, bool set)
 {
 	struct bch_fs *c = trans->c;
 
@@ -1383,14 +1394,14 @@ static int bch2_extent_reconcile_pending_mod(struct btree_trans *trans, struct b
 	struct bkey_i *n = errptr_try(bch2_trans_kmalloc(trans, bkey_bytes(k.k)));
 	bkey_reassemble(n, k);
 
-	if (!iter->min_depth) {
+	if (!level) {
 		bkey_reconcile_pending_mod(c, n, set);
 
 		return  bch2_trans_update(trans, iter, n, 0) ?:
 			bch2_trans_commit(trans, NULL, NULL,
 					  BCH_TRANS_COMMIT_no_enospc);
 	} else {
-		CLASS(btree_node_iter, iter2)(trans, iter->btree_id, k.k->p, 0, iter->min_depth - 1, 0);
+		CLASS(btree_node_iter, iter2)(trans, iter->btree_id, k.k->p, 0, level - 1, 0);
 		struct btree *b = errptr_try(bch2_btree_iter_peek_node(&iter2));
 
 		if (!bkey_and_val_eq(bkey_i_to_s_c(&b->key), bkey_i_to_s_c(n))) {
@@ -1408,39 +1419,93 @@ static int bch2_extent_reconcile_pending_mod(struct btree_trans *trans, struct b
 	}
 }
 
-static bool is_reconcile_pending_err(struct bch_fs *c, struct bkey_s_c k, int err)
+static int check_reconcile_pending_err(struct btree_trans *trans,
+				       struct bch_inode_opts *opts,
+				       struct data_update_opts *data_opts,
+				       struct bkey_s_c k, int err)
 {
-	 bool ret = (bch2_err_matches(err, BCH_ERR_data_update_fail_no_rw_devs) ||
-		     bch2_err_matches(err, BCH_ERR_insufficient_devices) ||
-		     bch2_err_matches(err, ENOSPC));
-	 if (ret)
-		event_add_trace(c, reconcile_set_pending, k.k->size, buf, ({
-			prt_printf(&buf, "%s\n", bch2_err_str(err));
-			bch2_bkey_val_to_text(&buf, c, k);
-		}));
-	 return ret;
+	struct bch_fs *c = trans->c;
+
+	 if (!bch2_err_matches(err, BCH_ERR_data_update_fail_no_rw_devs) &&
+	     !bch2_err_matches(err, BCH_ERR_insufficient_devices) &&
+	     !bch2_err_matches(err, ENOSPC))
+		 return err;
+
+	event_add_trace(c, reconcile_set_pending, k.k->size, buf, ({
+		prt_printf(&buf, "%s\n", bch2_err_str(err));
+		bch2_bkey_val_to_text(&buf, c, k);
+		prt_newline(&buf);
+		bch2_data_update_opts_to_text(&buf, c, opts, data_opts);
+		prt_newline(&buf);
+		int ret = bch2_can_do_data_update(trans, opts, data_opts, k, &buf);
+		if (bch2_err_matches(ret, BCH_ERR_transaction_restart))
+			return ret;
+	}));
+	return 1;
 }
 
 static int __do_reconcile_extent(struct moving_context *ctxt,
 				 struct per_snapshot_io_opts *snapshot_io_opts,
-				 struct btree_iter *iter, struct bkey_s_c k)
+				 struct bch_inode_opts *opts,
+				 struct data_update_opts *data_opts,
+				 struct bbpos work,
+				 struct btree_iter *iter,
+				 unsigned level,
+				 struct bkey_s_c k)
 {
+	if (!bkey_extent_is_direct_data(k.k))
+		return 0;
+
 	struct btree_trans *trans = ctxt->trans;
 	struct bch_fs *c = trans->c;
 	u32 restart_count = trans->restart_count;
 
 	ctxt->stats = &c->reconcile.work_stats;
 
-	int ret = bch2_move_extent(ctxt, NULL, snapshot_io_opts,
-				   reconcile_set_data_opts, NULL,
-				   iter, iter->min_depth, k);
+	try(bch2_bkey_get_io_opts(trans, snapshot_io_opts, k, opts));
+	try(bch2_update_reconcile_opts(trans, snapshot_io_opts, opts, iter, level, k,
+				       SET_NEEDS_REBALANCE_other));
+
+	CLASS(disk_reservation, res)(c);
+	try(bch2_trans_commit_lazy(trans, &res.r, NULL, BCH_TRANS_COMMIT_no_enospc));
+
+	int ret = reconcile_set_data_opts(trans, NULL, iter->btree_id, k, opts, data_opts);
+	if (ret <= 0)
+		return ret;
+
+	if (work.btree == BTREE_ID_reconcile_pending) {
+		int ret = bch2_can_do_data_update(trans, opts, data_opts, k, NULL);
+		ret = check_reconcile_pending_err(trans, opts, data_opts, k, ret);
+		if (ret > 0)
+			return 0;
+		if (ret)
+			return ret;
+
+		if (extent_has_rotational(c, k)) {
+			/*
+			 * The pending list is in logical inode:offset order,
+			 * but if the extent is on spinning rust we want do it
+			 * in device LBA order.
+			 *
+			 * Just take it off the pending list for now, and we'll
+			 * pick it up when we scan reconcile_work_phys:
+			 */
+			return bch2_extent_reconcile_pending_mod(trans, iter, level, k, false);
+		}
+	}
+
+	ret = bch2_move_extent(ctxt, NULL, opts, data_opts, iter, level, k);
+	BUG_ON(ret > 0);
+	ret = check_reconcile_pending_err(trans, opts, data_opts, k, ret);
+	if (ret > 0)
+		return bch2_extent_reconcile_pending_mod(trans, iter, level, k, true);
 	if (bch2_err_matches(ret, BCH_ERR_transaction_restart) ||
-	    bch2_err_matches(ret, EROFS))
+	    bch2_err_matches(ret, BCH_ERR_data_update_fail_need_copygc))
 		return ret;
-	if (is_reconcile_pending_err(c, k, ret))
-		return bch2_extent_reconcile_pending_mod(trans, iter, k, true);
 	if (ret) {
-		WARN_ONCE(ret != -BCH_ERR_data_update_fail_no_snapshot,
+		WARN_ONCE(!bch2_err_matches(ret, EROFS) &&
+			  ret != -BCH_ERR_data_update_fail_no_snapshot &&
+			  ret != -BCH_ERR_data_update_fail_in_flight,
 			  "unhandled error from move_extent: %s", bch2_err_str(ret));
 		/* skip it and continue */
 	}
@@ -1466,62 +1531,61 @@ static int do_reconcile_extent(struct moving_context *ctxt,
 	if (!k.k)
 		return 0;
 
-	if (work.btree == BTREE_ID_reconcile_pending) {
-		struct bch_inode_opts opts;
-		try(bch2_bkey_get_io_opts(trans, snapshot_io_opts, k, &opts));
+	struct bch_inode_opts opts;
+	struct data_update_opts data_opts = {};
+	try(__do_reconcile_extent(ctxt, snapshot_io_opts, &opts, &data_opts, work, &iter, 0, k));
 
-		struct data_update_opts data_opts = { .read_dev = -1 };
-		reconcile_set_data_opts(trans, NULL, data_pos.btree, k, &opts, &data_opts);
-
-		struct bch_devs_list devs_have = bch2_data_update_devs_keeping(c, &data_opts, k);
-		int ret = bch2_can_do_write(c, &opts, &data_opts, k, &devs_have);
-		if (ret) {
-			if (is_reconcile_pending_err(c, k, ret))
-				return 0;
-			return ret;
-		}
-
-		if (extent_has_rotational(c, k))
-			return bch2_extent_reconcile_pending_mod(trans, &iter, k, false);
-	}
-
-	event_add_trace(c, reconcile_data, k.k->size, buf,
-			bch2_bkey_val_to_text(&buf, c, k));
-
-	return __do_reconcile_extent(ctxt, snapshot_io_opts, &iter, k);
+	event_add_trace(c, reconcile_data, k.k->size, buf, ({
+		prt_newline(&buf);
+		bch2_bkey_val_to_text(&buf, c, k);
+		prt_newline(&buf);
+		bch2_data_update_opts_to_text(&buf, c, &opts, &data_opts);
+	}));
+	return 0;
 }
 
-static int do_reconcile_phys(struct moving_context *ctxt,
-			     struct per_snapshot_io_opts *snapshot_io_opts,
-			     struct bpos bp_pos, struct wb_maybe_flush *last_flushed)
+static int do_reconcile_extent_phys(struct moving_context *ctxt,
+				    struct per_snapshot_io_opts *snapshot_io_opts,
+				    struct bbpos work,
+				    struct wb_maybe_flush *last_flushed)
 {
 	struct btree_trans *trans = ctxt->trans;
 	struct bch_fs *c = trans->c;
 
-	CLASS(btree_iter, bp_iter)(trans, BTREE_ID_backpointers, bp_pos, 0);
+	CLASS(btree_iter, bp_iter)(trans, BTREE_ID_backpointers, work.pos, 0);
 	struct bkey_s_c bp_k = bkey_try(bch2_btree_iter_peek_slot(&bp_iter));
 	if (!bp_k.k || bp_k.k->type != KEY_TYPE_backpointer) /* write buffer race */
 		return 0;
 
 	CLASS(btree_iter_uninit, iter)(trans);
-	struct bkey_s_c k = bkey_try(bch2_backpointer_get_key(trans, bkey_s_c_to_backpointer(bp_k),
-							      &iter, 0, last_flushed));
+	struct bkey_s_c_backpointer bp = bkey_s_c_to_backpointer(bp_k);
+	struct bkey_s_c k = bkey_try(bch2_backpointer_get_key(trans, bp, &iter, 0, last_flushed));
 	if (!k.k)
 		return 0;
 
+	struct bch_inode_opts opts;
+	struct data_update_opts data_opts = {
+		.read_dev	= work.pos.inode,
+		.read_flags	= BCH_READ_soft_require_read_device,
+	};
+	try(__do_reconcile_extent(ctxt, snapshot_io_opts, &opts, &data_opts, work, &iter, bp.v->level, k));
+
 	event_add_trace(c, reconcile_phys, k.k->size, buf, ({
 		prt_newline(&buf);
 		bch2_bkey_val_to_text(&buf, c, bp_k);
 		prt_newline(&buf);
 		bch2_bkey_val_to_text(&buf, c, k);
+		prt_newline(&buf);
+		bch2_data_update_opts_to_text(&buf, c, &opts, &data_opts);
 	}));
 
-	return __do_reconcile_extent(ctxt, snapshot_io_opts, &iter, k);
+	return 0;
 }
 
 noinline_for_stack
 static int do_reconcile_btree(struct moving_context *ctxt,
 			      struct per_snapshot_io_opts *snapshot_io_opts,
+			      struct bbpos work,
 			      struct bkey_s_c_backpointer bp)
 {
 	struct btree_trans *trans = ctxt->trans;
@@ -1532,10 +1596,18 @@ static int do_reconcile_btree(struct moving_context *ctxt,
 	if (!k.k)
 		return 0;
 
-	event_add_trace(c, reconcile_btree, btree_sectors(c), buf,
-			bch2_bkey_val_to_text(&buf, c, k));
+	struct bch_inode_opts opts;
+	struct data_update_opts data_opts = {};
+	try(__do_reconcile_extent(ctxt, snapshot_io_opts, &opts, &data_opts, work, &iter, bp.v->level, k));
 
-	return __do_reconcile_extent(ctxt, snapshot_io_opts, &iter, k);
+	event_add_trace(c, reconcile_btree, btree_sectors(c), buf, ({
+		prt_newline(&buf);
+		bch2_bkey_val_to_text(&buf, c, k);
+		prt_newline(&buf);
+		bch2_data_update_opts_to_text(&buf, c, &opts, &data_opts);
+	}));
+
+	return 0;
 }
 
 static int update_reconcile_opts_scan(struct btree_trans *trans,
@@ -1595,19 +1667,13 @@ static int do_reconcile_scan_bps(struct moving_context *ctxt,
 
 	bch2_btree_write_buffer_flush_sync(trans);
 
-	CLASS(disk_reservation, res)(c);
+	return backpointer_scan_for_each(trans, iter, POS(s.dev, 0), POS(s.dev, U64_MAX),
+				  last_flushed, NULL, bp, ({
+		ctxt->stats->pos = BBPOS(BTREE_ID_backpointers, iter.pos);
 
-	return for_each_btree_key_max_commit(trans, iter, BTREE_ID_backpointers,
-					  POS(s.dev, 0), POS(s.dev, U64_MAX),
-					  BTREE_ITER_prefetch, k,
-					  &res.r, NULL, BCH_TRANS_COMMIT_no_enospc, ({
-		ctxt->stats->pos = BBPOS(iter.btree_id, iter.pos);
-
-		if (k.k->type != KEY_TYPE_backpointer)
-			continue;
-
-		bch2_disk_reservation_put(c, &res.r);
-		do_reconcile_scan_bp(trans, s, bkey_s_c_to_backpointer(k), last_flushed);
+		CLASS(disk_reservation, res)(c);
+		do_reconcile_scan_bp(trans, s, bp, last_flushed) ?:
+		bch2_trans_commit(trans, &res.r, NULL, BCH_TRANS_COMMIT_no_enospc);
 	}));
 }
 
@@ -1651,42 +1717,11 @@ static int do_reconcile_scan_btree(struct moving_context *ctxt,
 	struct bch_fs *c = trans->c;
 	struct bch_fs_reconcile *r = &c->reconcile;
 
-	/*
-	 * peek(), peek_slot() don't know how to fetch btree root keys - we
-	 * really should fix this
-	 */
-	while (level == bch2_btree_id_root(c, btree)->level + 1) {
-		bch2_trans_begin(trans);
-
-		CLASS(btree_node_iter, iter)(trans, btree, start, 0, level - 1,
-					     BTREE_ITER_prefetch|
-					     BTREE_ITER_not_extents|
-					     BTREE_ITER_all_snapshots);
-		struct btree *b = bch2_btree_iter_peek_node(&iter);
-		int ret = PTR_ERR_OR_ZERO(b);
-		if (ret)
-			goto root_err;
-
-		if (b != btree_node_root(c, b))
-			continue;
-
-		if (btree_node_fake(b))
-			return 0;
-
-		struct bkey_s_c k = bkey_i_to_s_c(&b->key);
-
+	try(for_btree_root_key_at_level(trans, iter, btree, level, k, ({
 		struct bch_inode_opts opts;
-		ret =   bch2_bkey_get_io_opts(trans, snapshot_io_opts, k, &opts) ?:
-			update_reconcile_opts_scan(trans, snapshot_io_opts, &opts, &iter, level, k, s);
-root_err:
-		if (bch2_err_matches(ret, BCH_ERR_transaction_restart))
-			continue;
-		if (bch2_err_matches(ret, BCH_ERR_data_update_fail))
-			ret = 0; /* failure for this extent, keep going */
-		WARN_ONCE(ret && !bch2_err_matches(ret, EROFS),
-			  "unhandled error from move_extent: %s", bch2_err_str(ret));
-		return ret;
-	}
+		bch2_bkey_get_io_opts(trans, snapshot_io_opts, k, &opts) ?:
+		update_reconcile_opts_scan(trans, snapshot_io_opts, &opts, &iter, level, k, s);
+	})));
 
 	bch2_trans_begin(trans);
 	CLASS(btree_node_iter, iter)(trans, btree, start, 0, level,
@@ -1807,11 +1842,91 @@ static void reconcile_wait(struct bch_fs *c)
 
 static bool bch2_reconcile_enabled(struct bch_fs *c)
 {
-	return c->opts.reconcile_enabled &&
+	return !c->opts.read_only &&
+		c->opts.reconcile_enabled &&
 		!(c->opts.reconcile_on_ac_only &&
 		  c->reconcile.on_battery);
 }
 
+typedef struct {
+	struct bch_fs		*c;
+	unsigned		dev;
+	enum btree_id		btree;
+	struct closure		cl;
+
+	struct bch_move_stats	stats;
+} reconcile_phys_thr;
+
+DEFINE_DARRAY(reconcile_phys_thr);
+
+static CLOSURE_CALLBACK(do_reconcile_phys_thread)
+{
+	closure_type(thr, reconcile_phys_thr, cl);
+	struct bch_fs *c = thr->c;
+
+	struct moving_context ctxt __cleanup(bch2_moving_ctxt_exit);
+	bch2_moving_ctxt_init(&ctxt, c, NULL, &thr->stats,
+			      writepoint_ptr(&c->allocator.reconcile_write_point),
+			      true);
+
+	struct btree_trans *trans = ctxt.trans;
+
+	CLASS(darray_reconcile_work, work)();
+	darray_make_room(&work, REBALANCE_WORK_BUF_NR);
+	if (!work.size) {
+		bch_err(c, "%s: unable to allocate memory", __func__);
+		closure_return(cl);
+		return;
+	}
+
+	CLASS(per_snapshot_io_opts, snapshot_io_opts)(c);
+
+	struct wb_maybe_flush last_flushed __cleanup(wb_maybe_flush_exit);
+	wb_maybe_flush_init(&last_flushed);
+
+	struct bbpos work_pos = BBPOS(thr->btree, POS(thr->dev, 0));
+
+	while (!bch2_move_ratelimit(&ctxt)) {
+		if (!bch2_reconcile_enabled(c))
+			break;
+
+		bch2_trans_begin(trans);
+
+		struct bkey_s_c k = next_reconcile_entry(trans, &work, thr->btree, &work_pos.pos);
+		if (bkey_err(k) ||
+		    !k.k ||
+		    k.k->p.inode != thr->dev)
+			break;
+
+		int ret = lockrestart_do(trans,
+			do_reconcile_extent_phys(&ctxt, &snapshot_io_opts, work_pos, &last_flushed));
+		if (ret)
+			break;
+	}
+
+	closure_return(cl);
+}
+
+static int do_reconcile_phys(struct bch_fs *c, enum btree_id btree)
+{
+	CLASS(darray_reconcile_phys_thr, thrs)();
+	CLASS(closure_stack, cl)();
+
+	for_each_member_device(c, ca)
+		if (ca->mi.rotational &&
+		    bch2_dev_is_online(ca))
+			try(darray_push(&thrs, ((reconcile_phys_thr) {
+						.c	= c,
+						.dev	= ca->dev_idx,
+						.btree	= btree,
+						})));
+
+	darray_for_each(thrs, i)
+		closure_call(&i->cl, do_reconcile_phys_thread, system_unbound_wq, &cl);
+
+	return 0;
+}
+
 static int do_reconcile(struct moving_context *ctxt)
 {
 	struct btree_trans *trans = ctxt->trans;
@@ -1819,6 +1934,7 @@ static int do_reconcile(struct moving_context *ctxt)
 	struct bch_fs_reconcile *r = &c->reconcile;
 	u64 sectors_scanned = 0;
 	u32 kick = r->kick;
+	u32 copygc_run_count = c->copygc.run_count;
 	int ret = 0;
 
 	CLASS(darray_reconcile_work, work)();
@@ -1888,24 +2004,6 @@ static int do_reconcile(struct moving_context *ctxt)
 			continue;
 		}
 
-		if ((r->work_pos.btree == BTREE_ID_reconcile_hipri_phys ||
-		     r->work_pos.btree == BTREE_ID_reconcile_work_phys) &&
-		    k.k->p.inode != r->work_pos.pos.inode) {
-			/*
-			 * We don't yet do multiple devices in parallel - that
-			 * will require extra synchronization to avoid kicking
-			 * off the same reconciles simultaneously via multiple
-			 * backpointers.
-			 *
-			 * For now, flush when switching devices to avoid
-			 * conflicts:
-			 */
-			bch2_moving_ctxt_flush_all(ctxt);
-			bch2_btree_write_buffer_flush_sync(trans);
-			work.nr = 0;
-			continue;
-		}
-
 		r->running = true;
 		r->work_pos.pos = k.k->p;
 
@@ -1926,15 +2024,26 @@ static int do_reconcile(struct moving_context *ctxt)
 			}
 
 			ret = do_reconcile_btree(ctxt, &snapshot_io_opts,
-						 bkey_s_c_to_backpointer(k));
+						 r->work_pos, bkey_s_c_to_backpointer(k));
 		} else if (btree_is_reconcile_phys(r->work_pos.btree)) {
-			ret = lockrestart_do(trans,
-				do_reconcile_phys(ctxt, &snapshot_io_opts, k.k->p, &last_flushed));
+			bch2_trans_unlock_long(trans);
+			ret = do_reconcile_phys(c, r->work_pos.btree);
+			r->work_pos = BBPOS(scan_btrees[++i], POS_MIN);
 		} else {
 			ret = lockrestart_do(trans,
 				do_reconcile_extent(ctxt, &snapshot_io_opts, r->work_pos));
 		}
 
+		if (bch2_err_matches(ret, BCH_ERR_data_update_fail_need_copygc)) {
+			bch2_trans_unlock_long(trans);
+			bch2_copygc_wakeup(c);
+			wait_event(c->copygc.running_wq,
+				   c->copygc.run_count != copygc_run_count ||
+				   kthread_should_stop());
+			ret = 0;
+			continue;
+		}
+
 		if (bch2_err_matches(ret, BCH_ERR_transaction_restart)) {
 			ret = 0;
 			continue;
@@ -2409,14 +2518,26 @@ static int check_reconcile_work_phys(struct btree_trans *trans)
 	}
 }
 
+static int btree_node_update_key_get_node(struct btree_trans *trans, struct btree_iter *iter,
+					  unsigned level, struct bkey_i *new_key)
+{
+	BUG_ON(!bkey_is_btree_ptr(&new_key->k));
+
+	CLASS(btree_node_iter, iter2)(trans, iter->btree_id, iter->pos, 0, level - 1, 0);
+	struct btree *b = errptr_try(bch2_btree_iter_peek_node(&iter2));
+
+	return bch2_btree_node_update_key(trans, &iter2, b, new_key, BCH_TRANS_COMMIT_no_enospc, false);
+}
+
 static int check_reconcile_work_btree_key(struct btree_trans *trans,
-					  struct btree_iter *iter, struct bkey_s_c k)
+					  struct btree_iter *iter,
+					  unsigned level, struct bkey_s_c k)
 {
 	struct bch_fs *c = trans->c;
 
 	struct bch_inode_opts opts;
 	try(bch2_bkey_get_io_opts(trans, NULL, k, &opts));
-	try(bch2_update_reconcile_opts(trans, NULL, &opts, iter, iter->min_depth, k,
+	try(bch2_update_reconcile_opts(trans, NULL, &opts, iter, level, k,
 				       SET_NEEDS_REBALANCE_other));
 
 	struct bpos bp_pos = bch2_bkey_get_reconcile_bp_pos(c, k);
@@ -2431,10 +2552,9 @@ static int check_reconcile_work_btree_key(struct btree_trans *trans,
 
 		bkey_reassemble(n, k);
 
-		try(reconcile_bp_add(trans, iter->btree_id, iter->min_depth,
-				     bkey_i_to_s(n), &bp_pos));
+		try(reconcile_bp_add(trans, iter->btree_id, level, bkey_i_to_s(n), &bp_pos));
 		bch2_bkey_set_reconcile_bp(c, bkey_i_to_s(n), bp_pos.offset);
-		return 0;
+		return btree_node_update_key_get_node(trans, iter, level, n);
 	}
 
 	if (ret_fsck_err_on(!bp_pos.inode && bp_pos.offset,
@@ -2446,15 +2566,14 @@ static int check_reconcile_work_btree_key(struct btree_trans *trans,
 		bkey_reassemble(n, k);
 		bch2_bkey_set_reconcile_bp(c, bkey_i_to_s(n), 0);
 
-		try(bch2_trans_update(trans, iter, n, BTREE_UPDATE_internal_snapshot_node));
-		return 0;
+		return btree_node_update_key_get_node(trans, iter, level, n);
 	}
 
 	if (!bpos_eq(bp_pos, POS_MIN)) {
 		CLASS(btree_iter, rb_iter)(trans, BTREE_ID_reconcile_scan, bp_pos, BTREE_ITER_intent);
 		struct bkey_s_c bp_k = bkey_try(bch2_btree_iter_peek_slot(&rb_iter));
 
-		struct bch_backpointer bp = rb_bp(iter->btree_id, iter->min_depth, k);
+		struct bch_backpointer bp = rb_bp(iter->btree_id, level, k);
 
 		if (bp_k.k->type != KEY_TYPE_backpointer || memcmp(bp_k.v, &bp, sizeof(bp))) {
 			CLASS(printbuf, buf)();
@@ -2475,8 +2594,12 @@ static int check_reconcile_work_btree_key(struct btree_trans *trans,
 				struct bkey_i_backpointer *new_bp = errptr_try(bch2_bkey_alloc(trans, &rb_iter, 0, backpointer));
 				new_bp->v = bp;
 
-				struct bkey_i *n = errptr_try(bch2_bkey_make_mut(trans, iter, &k, 0));
+				struct bkey_i *n = errptr_try(bch2_trans_kmalloc(trans, bkey_bytes(k.k)));
+				bkey_reassemble(n, k);
+
 				bch2_bkey_set_reconcile_bp(c, bkey_i_to_s(n), rb_iter.pos.offset);
+
+				return btree_node_update_key_get_node(trans, iter, level, n);
 			}
 		}
 	}
@@ -2498,6 +2621,13 @@ static int check_reconcile_work_btrees(struct btree_trans *trans)
 			continue;
 
 		for (unsigned level = 1; level < BTREE_MAX_DEPTH; level++) {
+			try(for_btree_root_key_at_level(trans, iter, btree, level, k, ({
+				BUG_ON(!bkey_is_btree_ptr(k.k));
+				check_reconcile_work_btree_key(trans, &iter, level, k) ?:
+				bch2_trans_commit(trans, &res.r, NULL, BCH_TRANS_COMMIT_no_enospc);
+			})));
+
+			bch2_trans_begin(trans);
 			CLASS(btree_node_iter, iter)(trans, btree, POS_MIN, 0, level,
 						     BTREE_ITER_prefetch|
 						     BTREE_ITER_not_extents|
@@ -2506,7 +2636,7 @@ static int check_reconcile_work_btrees(struct btree_trans *trans)
 			try(for_each_btree_key_continue(trans, iter, 0, k, ({
 				bch2_disk_reservation_put(c, &res.r);
 				bch2_progress_update_iter(trans, &progress, &iter) ?:
-				check_reconcile_work_btree_key(trans, &iter, k) ?:
+				check_reconcile_work_btree_key(trans, &iter, level, k) ?:
 				bch2_trans_commit(trans, &res.r, NULL, BCH_TRANS_COMMIT_no_enospc);
 			})));
 		}
